;;; cccp.composite-gates  -  database of composite gates
;;; Copyright (c) 2013, Masamitsu Bando
;;;
;;; Author: Masamitsu Bando <m.bando.phys@gmail.com>
;;; Version: 1.05
;;; URL: http://alice.math.kindai.ac.jp/~bando/
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the MIT License.
;;;
;;; Last modified: Thu 29 Aug 2013 03:27:33 JST

(define-module cccp.composite-gates
  (use math.const)
  (use util.list)
  (export-all))
(select-module cccp.composite-gates)

;; from left to right

(define (simple-setup gate) `(,gate))

;; SCROFULOUS
(define (scrof-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [th1 (if (= theta pi) pi (if (= theta (* 0.5 pi)) (/ (* pi 115.182) 180.0)))]
         [th2 pi]
         [ph1 (+ phi (acos (/ (* pi (cos th1))
                              (* -2.0 th1 (sin (* 0.5 theta))))))]
         [ph2 (- ph1 (acos (/ (* -0.5 pi) th1)))])
    `((,th1 . ,ph1) (,th2 . ,ph2) (,th1 . ,ph1))))

;; reduced SKinsC
(define (rskic-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [k (asin (* 0.5 (sin (* 0.5 theta))))]
         [th1 (- (* 0.5 theta) k)] [phm (+ phi pi)]
         [a (acos (/ (- (* 2.0 pi) theta) (* -4.0 pi)))]
         [thsk (- (* 2.0 pi) theta)] [thskp (* 2.0 pi)]
         [phsk1 (- (+ phi pi) a)] [phsk2 (+ (+ phi pi) a)]
         [thmul (+ th1 thsk)])
    `((,th1 . ,phi) (,thmul . ,phm)
      (,thskp . ,phsk1) (,thskp . ,phsk2)
      (,th1 . ,phm) (,th1 . ,phi))))

;; reduced CinBB
(define (rcib-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [wph_p (acos (/ theta (* -4.0 pi)))]
         [wph1 (+ wph_p phi)] [wph2 (+ phi (* 3.0 wph_p))]
         [wth1 pi] [wth2 (* 2.0 pi)]
         [k (asin (* 0.5 (sin (* 0.5 theta))))]
         [th1 (+ (* 2.0 pi) (* 0.5 theta) (- k))]
         [th2 (- (* 2.0 pi) (* 2 k))] [th3 (+ (* 0.5 theta) (- k))]
         [ph1 phi] [ph2 (+ phi pi)] [thx (* 2.0 pi)])
    `((,wth1 . ,wph1) (,wth2 . ,wph2) (,wth1 . ,wph1)
      (,th1 . ,ph1) (,th2 . ,ph2) (,th3 . ,ph1))))

(define (waltz-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [th1 (* 0.5 pi)] [th2 pi] [th3 (* 1.5 pi)] [th4 (* 2.0 pi)]
         [ph1 phi] [ph2 (+ phi pi)])
    `((,th1 . ,ph1) (,th2 . ,ph2) (,th4 . ,ph1) (,th2 . ,ph2) (,th3 . ,ph1)
      (,th1 . ,ph2) (,th2 . ,ph1) (,th4 . ,ph2) (,th2 . ,ph1) (,th3 . ,ph2))))

;; CORPSE
(define (corpse-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [k (asin (* 0.5 (sin (* 0.5 theta))))]
         [th1 (+ (* 2.0 pi) (* 0.5 theta) (- k))] [th2 (- (* 2.0 pi) (* 2 k))]
         [th3 (+ (* 0.5 theta) (- k))]
         [ph1 phi] [ph2 (+ phi pi)])
    `((,th1 . ,ph1) (,th2 . ,ph2) (,th3 . ,ph1))))

;; short CORPSE
(define (scorpse-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [k (asin (* 0.5 (sin (* 0.5 theta))))]
         [th1 (- (* 0.5 theta) k)] [th2 (- (* 2 pi) (* 2 k))]
         [ph1 phi] [ph2 (+ phi pi)])
    `((,th1 . ,ph1) (,th2 . ,ph2) (,th1 . ,ph1))))

;; for concatenation
;; CORPSE (n_1 = n_3 = 1, n_2 = 2)
(define (corpsex-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [k (asin (* 0.5 (sin (* 0.5 theta))))]
         [th1 (+ (* 2.0 pi) (* 0.5 theta) (- k))]
         [th2 (- (* 4.0 pi) (* 2 k))]
         [ph1 phi] [ph2 (+ phi pi)]
         [thx (* 2.0 pi)])
    `((,th1 . ,ph1) (,th2 . ,ph2) (,th1 . ,ph1))))

;; if (theta = 2*PI) { gate } else { CORPSE }
(define (corpseg-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [k (asin (* 0.5 (sin (* 0.5 theta))))]
         [th1 (+ (* 2.0 pi) (* 0.5 theta) (- k))] [th2 (- (* 2.0 pi) (* 2 k))]
         [th3 (+ (* 0.5 theta) (- k))]
         [ph1 phi] [ph2 (+ phi pi)]
         [thx (* 2.0 pi)])
    (if (<= (mod theta (* 2 pi)) 1e-8)
        `(,gate)
        `((,th1 . ,ph1) (,th2 . ,ph2) (,th3 . ,ph1)))))

;; if (theta = 2*PI) { gate } else { corpsex }
(define (corpsexg-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [k (asin (* 0.5 (sin (* 0.5 theta))))]
         [th1 (+ (* 2.0 pi) (* 0.5 theta) (- k))]
         [th2 (- (* 4.0 pi) (* 2 k))]
         [ph1 phi] [ph2 (+ phi pi)]
         [thx (* 2.0 pi)])
    (if (<= (mod theta (* 2 pi)) 1e-8)
        `(,gate)
        `((,th1 . ,ph1) (,th2 . ,ph2) (,th1 . ,ph1)))))

;; SK1
(define (sk-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [a (acos (/ theta (* -4.0 pi)))]
         [ph1 (- phi a)] [ph2 (+ phi a)]
         [th1 (* 2.0 pi)])
    `((,theta . ,phi) (,th1 . ,ph1) (,th1 . ,ph2))))

;; symmetric SK1
(define (sks-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [a (acos (/ theta (* -8.0 pi)))]
         [ph1 (+ phi a)] [ph2 (- phi a)]
         [th1 (* 2.0 pi)])
    `((,th1 . ,ph1) (,th1 . ,ph2) (,theta . ,phi) (,th1 . ,ph2)
      (,th1 . ,ph1))))

;; symmetric BB1
(define (w1-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [th1 (* 0.5 theta)] [ph1 phi]
         [wph_p (acos (/ theta (* -4.0 pi)))]
         [wph1 (+ wph_p phi)] [wph2 (+ phi (* 3.0 wph_p))]
         [wth1 pi] [wth2 (* 2.0 pi)])
    `((,th1 . ,ph1) (,wth1 . ,wph1) (,wth2 . ,wph2) (,wth1 . ,wph1)
      (,th1 . ,ph1))))

;; BB1
(define (bb1-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [wph_p (acos (/ theta (* -4.0 pi)))]
         [wph1 (+ wph_p phi)] [wph2 (+ phi (* 3.0 wph_p))]
         [wth1 pi] [wth2 (* 2.0 pi)])
    `((,wth1 . ,wph1) (,wth2 . ,wph2) (,wth1 . ,wph1)
      (,theta . ,phi))))

;; novel pulse for ORE
(define (novel-setup gate)
  (let* ([theta (car gate)] [phi (cdr gate)]
         [th2 (* 2.0 (asin (/ (sin (* 0.5 theta)) (sqrt 2))))]
         [th1 (- (* 2.0 pi) (asin (tan (* 0.5 th2))))]
         [ph2 (+ phi (* 0.25 pi))] [ph1 (+ ph2 (* 0.5 pi))])
    `((,th1 . ,ph1) (,th2 . ,ph2) (,th1 . ,ph1))))


(define composite-alist
  `(("simple" . ,simple-setup)
    ("corpse" . ,corpse-setup)
    ("scrof" . ,scrof-setup)
    ("sk" . ,sk-setup)
    ("sks" . ,sks-setup)
    ("w1c" . ,w1-setup)
    ("b1c" . ,bb1-setup)
    ("novel" . ,novel-setup)
    ;; for concatenation
    ("cor" . ,corpse-setup)
    ("scor" . ,scorpse-setup)
    ("corx" . ,corpsex-setup)
    ("corg" . ,corpseg-setup)
    ("corxg" . ,corpsexg-setup)
    ("scr" . ,scrof-setup)
    ("nov" . ,novel-setup)
    ("rskic" . ,rskic-setup)
    ("rcib" . ,rcib-setup)
    ("waltz" . ,waltz-setup)
    ))

(define (composite-setup gate-type gate)
  (let ([stype-list (string-split gate-type #/_in_|\//)]
        [ap-gate (lambda (s) (assoc-ref composite-alist s))])
    (let loop ([l (reverse stype-list)] [gate-l (list gate)])
      (if (null? l) gate-l
          (loop (cdr l) (apply append (map (ap-gate (car l)) gate-l)))))))

(provide "cccp/composite-gates")
